
/* 
 * Take co-registered TM and DEM data in *binary* format, 
 * read in a specified subsection of that data, and write it to files
 * as both a binary PPM file (output_file_name.ppm) and an ASCII file 
 * containing elevation and color data (output_file_name.data).
 *
 * CHANGES: The output file is now not PPM, it is split into a PPM header
 * file (output_file_name.header) and an RGB data file (output_file_name.rgb).
 * If you concatenate these you will get a standard PPM file.
 *
 * Program is called with 5 input variables 
 *   <DEM data input file> 
 *   <TM data file for RED> 
 *   <TM data file for GREEN> 
 *   <TM data file for BLUE> 
 *   <output_file_name>
 * and the coordinates of the area from stdin
 * 
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

long  North = 4792877;
long  South = 4692877;
long  East = 631732;
long  West = 531732;
long  M = 3333;
long  N = 3333;

float RedScale   = 3.0;   /* Scale the raw RBG colors */
float GreenScale = 3.5;
float BlueScale  = 2.0;

int Cutoff = 220;

int GridSpacing = 30;

/* Coordinates of subset of data to be output */
int x0, x1, y0, y1;

main(int argc, char *argv[])
{
   FILE *OutputData;
   int   InputDEM;
   int   InputRED;
   int   InputGREEN;
   int   InputBLUE;
   int   OutputPPM;
   char  Header[256];
   char  filename[256];
   char  line[4096];
   short  DEM;
   short *DEM_Buffer;
   unsigned char  *Color_Buffer;
   unsigned char  *RED_Buffer;
   unsigned char  *GREEN_Buffer;
   unsigned char  *BLUE_Buffer;
   long   DEM_Buffer_Size;
   long   Color_Buffer_Size;
   long   TM_Buffer_Size;
   int    i,j, tmp;
   int    Index;
   
   if (argc < 5)
   {
      printf("usage: <GRASS DEM datafile> <RED TM datafile> \
      <GREEN TM datafile> <BLUE TM datafile> <output file name>\n");
      exit(1);
   }
   
   /* Open input file of binary DEM data */
   sprintf(filename, "%s", argv[1]);
   if ((InputDEM = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   /* Open input file of RED band TM data */
   sprintf(filename, "%s", argv[2]);
   if ((InputRED = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   /* Open input file of GREEN band TM data */
   sprintf(filename, "%s", argv[3]);
   if ((InputGREEN = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   /* Open input file of BLUE band TM data */
   sprintf(filename, "%s", argv[4]);
   if ((InputBLUE = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }


   /* Read in the requested data coordinates */

   printf("The data covers an area of %ld x %ld grid points, \n", M, N);
   printf(" with grid points %d meters apart.\n\n", GridSpacing);
   printf("The area has UTM coordinates (in meters) of \n"); 
   printf(" (North, South, East, West) = (%ld, %ld, %ld, %ld)\n\n",
           North, South, East, West);

   printf("Enter the (N, S, E, W) co-ordinates of the section of this area\n");
   printf("you wish to display, in units of grid points\n");
   printf("(0 <= S <= N <= %ld, 0 <= W <= E <= %ld) \n", M, N);

   printf("\n South? "); scanf("%d", &x0);
   printf("\n North? "); scanf("%d", &x1);
   printf("\n West?  "); scanf("%d", &y0);
   printf("\n East?  "); scanf("%d", &y1);
   printf("\n");

   /* Fix up bogus data requests */

   if ((x0 > M) || (x1 > M)) {
      printf("North - South coordinates must be less than %ld\n", M);
      fflush(stdout);
      exit(4);
   }
   if ((y0 > N) || (y1 > N)) {
      printf("East - West coordinates must be less than %ld\n", N);
      fflush(stdout);
      exit(5);
   }
   if (x0 > x1) {
     tmp = x0; x0 = x1; x1 = tmp;
      printf("Changed input values so that North > South\n");
   }
   if (y0 > y1) {
     tmp = y0; y0 = y1; y1 = tmp;
      printf("Changed input values so that East > West\n");
   }

   printf("South = %d  North = %d  West = %d  East = %d\n\n", x0, x1, y0, y1);
   
   /* Open output file for DEM and color data */
   sprintf(filename, "%s.data", argv[5]);
   if ((OutputData = fopen(filename, "w")) == NULL)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   fprintf(OutputData, "north: %d\n", South + (x1*GridSpacing)); 
   fprintf(OutputData, "south: %d\n", South + (x0*GridSpacing)); 
   fprintf(OutputData, "east: %d\n", West + (y1*GridSpacing)); 
   fprintf(OutputData, "west: %d\n", West + (y0*GridSpacing)); 
   fprintf(OutputData, "rows: %d\n", x1-x0);
   fprintf(OutputData, "cols: %d\n", y1-y0);


   /* Set up data arrays */

   DEM_Buffer_Size = N * sizeof(short);
   TM_Buffer_Size = N * sizeof(unsigned char);
   Color_Buffer_Size = 3 * (y1-y0) * sizeof(unsigned char);

   if ((DEM_Buffer = (short *) malloc(DEM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((RED_Buffer = (unsigned char *) malloc(TM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((GREEN_Buffer = (unsigned char *) malloc(TM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((BLUE_Buffer = (unsigned char *) malloc(TM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((Color_Buffer = (unsigned char *) malloc(Color_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   /* Open output PPM header file for processed data */
   sprintf(filename, "%s.header", argv[5]);
   if ((OutputPPM = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   sprintf(Header, "P6\n%d %d\n255\n\0", y1-y0,  x1-x0);
   write(OutputPPM, Header, strlen(Header));
   close(OutputPPM);
 

   /* Open output coordinate header file for processed data */
   sprintf(filename, "%s.coord", argv[5]);
   if ((OutputPPM = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   write(OutputPPM, &North, sizeof(long));
   write(OutputPPM, &South, sizeof(long));
   write(OutputPPM, &East,  sizeof(long));
   write(OutputPPM, &West,  sizeof(long));
   write(OutputPPM, &M, sizeof(long));
   write(OutputPPM, &N, sizeof(long));
   close(OutputPPM);

   /* Open output PPM image file for processed data */
   sprintf(filename, "%s.rgb", argv[5]);
   if ((OutputPPM = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }

   /*  Read in and process the DEM and TM data */
   /*  N.B. Data is ordered from top left to bottom right, so need to  *
    *  count down from the largest value of i (i.e. North to South)    */

   printf("Reading in and processing the DEM and TM data ....\n");
   fflush(stdout);

   for (i = M-1; i >= x0; i--)
   {

      /* Read in rows of binary DEM and TM data */
      read(InputDEM,   DEM_Buffer,   DEM_Buffer_Size);
      read(InputRED,   RED_Buffer,   TM_Buffer_Size);
      read(InputGREEN, GREEN_Buffer, TM_Buffer_Size);
      read(InputBLUE,  BLUE_Buffer,  TM_Buffer_Size);
	
      if ((i % 100) == 0) printf("row %d\n", i);
      
      /* Only process data if it is in the requested area */
      if (i < x1) {
      
        Index = 0;

        for (j = y0; j < y1; j++)
        {
	   /* Convert to something approximating true color */
	   Color_Buffer[Index]
	      = (unsigned char)(RedScale   * RED_Buffer[j]);
           Color_Buffer[Index + 1]
	      = (unsigned char)(GreenScale * GREEN_Buffer[j]);
           Color_Buffer[Index + 2]
	      = (unsigned char)(BlueScale  * BLUE_Buffer[j]);

	   if (Color_Buffer[Index]   > Cutoff) Color_Buffer[Index]   = Cutoff;
	   if (Color_Buffer[Index+1] > Cutoff) Color_Buffer[Index+1] = Cutoff;
	   if (Color_Buffer[Index+2] > Cutoff) Color_Buffer[Index+2] = Cutoff;

	   fprintf(OutputData, "%hd %d %d %d\n", 
				    DEM_Buffer[j],
		               (int)Color_Buffer[Index], 
		               (int)Color_Buffer[Index+1], 
		               (int)Color_Buffer[Index+2]);
	   Index += 3;
        }

        /* Print out row of color PPM file */
        write(OutputPPM, Color_Buffer, Color_Buffer_Size);

      }
   }

   close(InputDEM);
   close(InputRED);
   close(InputGREEN);
   close(InputBLUE);
   close(OutputPPM);
   fclose(OutputData);
}

